use serde::{Deserialize, Serialize};

use super::berries::*;
use super::evolution::*;
use super::games::*;
use super::get_api_loc_from_url;
use super::items::*;
use super::locations::*;
use super::moves::*;
use super::resource_lists::*;
use super::utility::*;

use crate::cache::get_resource;
use crate::{impl_id, impl_id_and_named, set_endpoint};

/// <https://pokeapi.co/docs/v2.html#abilities>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct Ability {
    pub id: i16,
    pub name: String,
    pub is_main_series: bool,
    pub generation: NamedAPIResource<Generation>,
    pub names: Vec<Name>,
    pub effect_entries: Vec<VerboseEffect>,
    pub effect_changes: Vec<AbilityEffectChange>,
    pub flavor_text_entries: Vec<AbilityFlavorText>,
    pub pokemon: Vec<AbilityPokemon>,
}

/// <https://pokeapi.co/docs/v2.html#abilityeffectchange>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct AbilityEffectChange {
    pub effect_entries: Vec<Effect>,
    pub version_group: NamedAPIResource<VersionGroup>,
}

/// <https://pokeapi.co/docs/v2.html#abilityflavortext>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct AbilityFlavorText {
    pub flavor_text: String,
    pub language: NamedAPIResource<Language>,
    pub version_group: NamedAPIResource<VersionGroup>,
}

/// <https://pokeapi.co/docs/v2.html#abilitypokemon>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct AbilityPokemon {
    pub is_hidden: bool,
    pub slot: u8,
    pub pokemon: NamedAPIResource<Pokemon>,
}

/// <https://pokeapi.co/docs/v2.html#characteristics>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct Characteristic {
    pub id: i16,
    pub gene_modulo: i8,
    pub possible_values: Vec<u8>,
    pub highest_stat: NamedAPIResource<Stat>, // not documented
    pub descriptions: Vec<Description>,       // not documented
}

/// <https://pokeapi.co/docs/v2.html#egg-groups>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct EggGroup {
    pub id: i16,
    pub name: String,
    pub names: Vec<Name>,
    pub pokemon_species: Vec<NamedAPIResource<PokemonSpecies>>,
}

/// <https://pokeapi.co/docs/v2.html#genders>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct Gender {
    pub id: i16,
    pub name: String,
    pub pokemon_species_details: Vec<PokemonSpeciesGender>,
    pub required_for_evolution: Vec<NamedAPIResource<PokemonSpecies>>,
}

/// <https://pokeapi.co/docs/v2.html#pokemonspeciesgender>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonSpeciesGender {
    pub rate: i8,
    pub pokemon_species: NamedAPIResource<PokemonSpecies>,
}

/// <https://pokeapi.co/docs/v2.html#growth-rates>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct GrowthRate {
    pub id: i16,
    pub name: String,
    pub formula: String,
    pub descriptions: Vec<Description>,
    pub levels: Vec<GrowthRateExperienceLevel>,
    pub pokemon_species: Vec<NamedAPIResource<PokemonSpecies>>,
}

/// <https://pokeapi.co/docs/v2.html#growthrateexperiencelevel>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct GrowthRateExperienceLevel {
    pub level: u8,
    pub experience: u32,
}

/// <https://pokeapi.co/docs/v2.html#natures>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct Nature {
    pub id: i16,
    pub name: String,
    pub decreased_stat: Option<NamedAPIResource<Stat>>,
    pub increased_stat: Option<NamedAPIResource<Stat>>,
    pub hates_flavor: Option<NamedAPIResource<BerryFlavor>>,
    pub likes_flavor: Option<NamedAPIResource<BerryFlavor>>,
    pub pokeathlon_stat_changes: Vec<NatureStatChange>,
    pub move_battle_style_preferences: Vec<MoveBattleStylePreference>,
    pub names: Vec<Name>,
}

/// <https://pokeapi.co/docs/v2.html#naturestatchange>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct NatureStatChange {
    pub max_change: i8,
    pub pokeathlon_stat: NamedAPIResource<PokeathlonStat>,
}

/// <https://pokeapi.co/docs/v2.html#movebattlestylepreference>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct MoveBattleStylePreference {
    pub low_hp_preference: u8,
    pub high_hp_preference: u8,
    pub move_battle_style: NamedAPIResource<MoveBattleStyle>,
}

/// <https://pokeapi.co/docs/v2.html#pokeathlon-stats>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokeathlonStat {
    pub id: i16,
    pub name: String,
    pub names: Vec<Name>,
    pub affecting_natures: NaturePokeathlonStatAffectSets,
}

/// <https://pokeapi.co/docs/v2.html#naturepokeathlonstataffectsets>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct NaturePokeathlonStatAffectSets {
    pub increase: Vec<NaturePokeathlonStatAffect>,
    pub decrease: Vec<NaturePokeathlonStatAffect>,
}

/// <https://pokeapi.co/docs/v2.html#naturepokeathlonstataffect>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct NaturePokeathlonStatAffect {
    pub max_change: i8,
    pub nature: NamedAPIResource<Nature>,
}

/// <https://pokeapi.co/docs/v2.html#pokemon>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct Pokemon {
    pub id: i16,
    pub name: String,
    pub base_experience: Option<u16>,
    pub height: u8,
    pub is_default: bool,
    pub order: i16,
    pub weight: u16,
    pub abilities: Vec<PokemonAbility>,
    pub forms: Vec<NamedAPIResource<PokemonForm>>,
    pub game_indices: Vec<VersionGameIndex>,
    pub held_items: Vec<PokemonHeldItem>,
    pub location_area_encounters: String, // TODO: implement a way to retrieve these
    pub moves: Vec<PokemonMove>,
    pub sprites: PokemonSprites,
    pub species: NamedAPIResource<PokemonSpecies>,
    pub stats: Vec<PokemonStat>,
    pub types: Vec<PokemonType>,
    pub past_types: Vec<PokemonPastType>,
    #[serde(default)] 
    pub detail: String,
}

impl Pokemon {
    /// Fetch list of `LocationAreaEncounters` from the API for this Pokemon.
    pub fn get_encounters(&self) -> Result<Vec<LocationAreaEncounter>, minreq::Error> {
        let loc = get_api_loc_from_url(&self.location_area_encounters);
        get_resource(loc)?.json::<Vec<LocationAreaEncounter>>()
    }
}

/// <https://pokeapi.co/docs/v2.html#pokemonability>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonAbility {
    pub is_hidden: bool,
    pub slot: u8,
    pub ability: NamedAPIResource<Ability>,
}

/// <https://pokeapi.co/docs/v2.html#pokemontype>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonType {
    pub slot: u8,
    #[serde(rename = "type")]
    pub type_: NamedAPIResource<Type>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonPastType {
    pub generation: NamedAPIResource<Generation>,
    pub types: Vec<PokemonType>,
}

/// <https://pokeapi.co/docs/v2.html#pokemonhelditem>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonHeldItem {
    pub item: NamedAPIResource<Item>,
    pub version_details: Vec<PokemonHeldItemVersion>,
}

/// <https://pokeapi.co/docs/v2.html#pokemonhelditemversion>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonHeldItemVersion {
    pub version: NamedAPIResource<Version>,
    pub rarity: u8,
}

/// <https://pokeapi.co/docs/v2.html#pokemonmove>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonMove {
    #[serde(rename = "move")]
    pub move_: NamedAPIResource<Move>,
    pub version_group_details: Vec<PokemonMoveVersion>,
}

/// <https://pokeapi.co/docs/v2.html#pokemonmoveversion>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonMoveVersion {
    pub move_learn_method: NamedAPIResource<MoveLearnMethod>,
    pub version_group: NamedAPIResource<VersionGroup>,
    pub level_learned_at: u8,
}

/// <https://pokeapi.co/docs/v2.html#pokemonstat>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonStat {
    pub stat: NamedAPIResource<Stat>,
    pub effort: u8,
    pub base_stat: u8,
}

/// <https://pokeapi.co/docs/v2.html#pokemonsprites>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonSprites {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny_female: Option<String>,
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub back_female: Option<String>,
    pub back_shiny_female: Option<String>,
    pub other: ArtOther,
    pub versions: ArtVersions,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtDreamWorld {
    pub front_default: Option<String>,
    pub front_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtOfficial {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtHome {
    pub front_default: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny: Option<String>,
    pub front_shiny_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtOther {
    pub dream_world: ArtDreamWorld,
    #[serde(rename = "official-artwork")]
    pub official_artwork: ArtOfficial,
    pub home: ArtHome,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtRedBlue {
    pub back_default: Option<String>,
    pub back_gray: Option<String>,
    pub back_transparent: Option<String>,
    pub front_default: Option<String>,
    pub front_gray: Option<String>,
    pub front_transparent: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtYellow {
    pub back_default: Option<String>,
    pub back_gray: Option<String>,
    pub back_transparent: Option<String>,
    pub front_default: Option<String>,
    pub front_gray: Option<String>,
    pub front_transparent: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtGenI {
    #[serde(rename = "red-blue")]
    pub red_blue: ArtRedBlue,
    pub yellow: ArtYellow,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtCrystal {
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub back_shiny_transparent: Option<String>,
    pub back_transparent: Option<String>,
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_shiny_transparent: Option<String>,
    pub front_transparent: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtGold {
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_transparent: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtSilver {
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_transparent: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtGenII {
    pub crystal: ArtCrystal,
    pub gold: ArtGold,
    pub silver: ArtSilver,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtEmerald {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtFRLG {
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtRubySapphire {
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtGenIII {
    pub emerald: ArtEmerald,
    #[serde(rename = "firered-leafgreen")]
    pub firered_leafgreen: ArtFRLG,
    #[serde(rename = "ruby-sapphire")]
    pub ruby_sapphire: ArtRubySapphire,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtDiamondPearl {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny_female: Option<String>,
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub back_female: Option<String>,
    pub back_shiny_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtHGSS {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny_female: Option<String>,
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub back_female: Option<String>,
    pub back_shiny_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtPlatinum {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny_female: Option<String>,
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub back_female: Option<String>,
    pub back_shiny_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtGenIV {
    #[serde(rename = "diamond-pearl")]
    pub diamond_pearl: ArtDiamondPearl,
    #[serde(rename = "heartgold-soulsilver")]
    pub heartgold_soulsilver: ArtHGSS,
    pub platinum: ArtPlatinum,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtAnimated {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny_female: Option<String>,
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub back_female: Option<String>,
    pub back_shiny_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtBlackWhite {
    pub animated: ArtAnimated,
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny_female: Option<String>,
    pub back_default: Option<String>,
    pub back_shiny: Option<String>,
    pub back_female: Option<String>,
    pub back_shiny_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtGenV {
    #[serde(rename = "black-white")]
    pub black_white: ArtBlackWhite,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtORAS {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtXY {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtGenVI {
    #[serde(rename = "omegaruby-alphasapphire")]
    pub omegaruby_alphasapphire: ArtORAS,
    #[serde(rename = "x-y")]
    pub x_y: ArtXY,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtUSUM {
    pub front_default: Option<String>,
    pub front_shiny: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtIcons {
    pub front_default: Option<String>,
    pub front_female: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtGenVII {
    pub icons: Option<ArtIcons>,
    #[serde(rename = "ultra-sun-ultra-moon")]
    pub ultrasun_ultramoon: ArtUSUM,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtGenVIII {
    pub icons: ArtIcons,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct ArtVersions {
    #[serde(rename = "generation-i")]
    pub generation_i: ArtGenI,
    #[serde(rename = "generation-ii")]
    pub generation_ii: ArtGenII,
    #[serde(rename = "generation-iii")]
    pub generation_iii: ArtGenIII,
    #[serde(rename = "generation-iv")]
    pub generation_iv: ArtGenIV,
    #[serde(rename = "generation-v")]
    pub generation_v: ArtGenV,
    #[serde(rename = "generation-vi")]
    pub generation_vi: ArtGenVI,
    #[serde(rename = "generation-vii")]
    pub generation_vii: ArtGenVII,
    #[serde(rename = "generation-viii")]
    pub generation_viii: Option<ArtGenVIII>,
}

/// <https://pokeapi.co/docs/v2.html#locationareaencounter>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct LocationAreaEncounter {
    pub location_area: NamedAPIResource<LocationArea>,
    pub version_details: Vec<VersionEncounterDetail>,
}

/// <https://pokeapi.co/docs/v2.html#pokemon-colors>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonColor {
    pub id: i16,
    pub name: String,
    pub names: Vec<Name>,
    pub pokemon_species: Vec<NamedAPIResource<PokemonSpecies>>,
}

/// <https://pokeapi.co/docs/v2.html#pokemon-forms>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonForm {
    pub id: i16,
    pub name: String,
    pub order: i16,
    pub form_order: u16,
    pub is_default: bool,
    pub is_battle_only: bool,
    pub is_mega: bool,
    pub form_name: String,
    pub pokemon: NamedAPIResource<Pokemon>,
    pub sprites: PokemonFormSprites,
    pub version_group: NamedAPIResource<VersionGroup>,
    pub names: Vec<Name>,
    pub form_names: Vec<Name>,
    pub types: Vec<PokemonType>,
    #[serde(default)] 
    pub detail: String,
}

/// <https://pokeapi.co/docs/v2.html#pokemonformsprites>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonFormSprites {
    pub front_default: Option<String>,
    pub front_female: Option<String>,
    pub front_shiny: Option<String>,
    pub front_shiny_female: Option<String>,
    pub back_default: Option<String>,
    pub back_female: Option<String>,
    pub back_shiny: Option<String>,
    pub back_shiny_female: Option<String>,
}

/// <https://pokeapi.co/docs/v2.html#pokemon-habitats>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonHabitat {
    pub id: i16,
    pub name: String,
    pub names: Vec<Name>,
    pub pokemon_species: Vec<NamedAPIResource<PokemonSpecies>>,
}

/// <https://pokeapi.co/docs/v2.html#pokemon-shapes>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonShape {
    pub id: i16,
    pub name: String,
    pub awesome_names: Vec<AwesomeName>,
    pub names: Vec<Name>,
    pub pokemon_species: Vec<NamedAPIResource<PokemonSpecies>>, // incorrectly documented as list PokemonSpecies
}

/// <https://pokeapi.co/docs/v2.html#awesomename>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct AwesomeName {
    pub awesome_name: String,
    pub language: NamedAPIResource<Language>,
}

/// <https://pokeapi.co/docs/v2.html#pokemon-species>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonSpecies {
    pub id: i16,
    pub name: String,
    pub order: Option<i16>,
    pub gender_rate: i8,
    pub capture_rate: u8,
    pub base_happiness: Option<u8>,
    pub is_baby: bool,
    pub is_legendary: bool,
    pub is_mythical: bool,
    pub hatch_counter: u8,
    pub has_gender_differences: bool,
    pub forms_switchable: bool,
    pub growth_rate: Option<NamedAPIResource<GrowthRate>>,
    pub pokedex_numbers: Vec<PokemonSpeciesDexEntry>,
    pub egg_groups: Vec<NamedAPIResource<EggGroup>>,
    pub color: Option<NamedAPIResource<PokemonColor>>,
    pub shape: Option<NamedAPIResource<PokemonShape>>,
    pub evolves_from_species: Option<NamedAPIResource<PokemonSpecies>>,
    pub evolution_chain: APIResource<EvolutionChain>,
    pub habitat: Option<NamedAPIResource<PokemonHabitat>>,
    pub generation: NamedAPIResource<Generation>,
    pub names: Vec<Name>,
    pub pal_park_encounters: Vec<PalParkEncounterArea>,
    pub flavor_text_entries: Vec<FlavorText>,
    pub form_descriptions: Vec<Description>,
    pub genera: Vec<Genus>,
    pub varieties: Vec<PokemonSpeciesVariety>,
    #[serde(default)] 
    pub detail: String,
}

/// <https://pokeapi.co/docs/v2.html#genus>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct Genus {
    pub genus: String,
    pub language: NamedAPIResource<Language>,
}

/// <https://pokeapi.co/docs/v2.html#pokemonspeciesdexentry>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonSpeciesDexEntry {
    pub entry_number: u16,
    pub pokedex: NamedAPIResource<Pokedex>,
}

/// <https://pokeapi.co/docs/v2.html#palparkencounterarea>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PalParkEncounterArea {
    pub base_score: u8,
    pub rate: u8,
    pub area: NamedAPIResource<PalParkArea>,
}

/// <https://pokeapi.co/docs/v2.html#pokemonspeciesvariety>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PokemonSpeciesVariety {
    pub is_default: bool,
    pub pokemon: NamedAPIResource<Pokemon>,
}

/// <https://pokeapi.co/docs/v2.html#stats>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct Stat {
    pub id: i16,
    pub name: String,
    pub game_index: i16,
    pub is_battle_only: bool,
    pub affecting_moves: MoveStatAffectSets,
    pub affecting_natures: NatureStatAffectSets,
    pub characteristics: Vec<APIResource<Characteristic>>, // incorrectly documented as APIResource
    pub move_damage_class: Option<NamedAPIResource<MoveDamageClass>>,
    pub names: Vec<Name>,
}

/// <https://pokeapi.co/docs/v2.html#movestataffectsets>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct MoveStatAffectSets {
    pub increase: Vec<MoveStatAffect>,
    pub decrease: Vec<MoveStatAffect>,
}

/// <https://pokeapi.co/docs/v2.html#movestataffect>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct MoveStatAffect {
    pub change: i8,
    #[serde(rename = "move")]
    pub move_: NamedAPIResource<Move>,
}

/// <https://pokeapi.co/docs/v2.html#naturestataffectsets>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct NatureStatAffectSets {
    pub increase: Vec<NamedAPIResource<MoveStatAffect>>,
    pub decrease: Vec<NamedAPIResource<MoveStatAffect>>,
}

/// <https://pokeapi.co/docs/v2.html#types>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct Type {
    pub id: i16,
    pub name: String,
    pub damage_relations: TypeRelations,
    pub game_indices: Vec<GenerationGameIndex>,
    pub generation: NamedAPIResource<Generation>,
    pub move_damage_class: Option<NamedAPIResource<MoveDamageClass>>,
    pub names: Vec<Name>,
    pub pokemon: Vec<TypePokemon>,
    pub moves: Vec<NamedAPIResource<Move>>,
    pub past_damage_relations: Vec<PastTypeRelations>,
}

/// <https://pokeapi.co/docs/v2.html#typepokemon>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct TypePokemon {
    pub slot: u8,
    pub pokemon: NamedAPIResource<Pokemon>,
}

/// <https://pokeapi.co/docs/v2.html#typerelations>
#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct TypeRelations {
    pub no_damage_to: Vec<NamedAPIResource<Type>>,
    pub half_damage_to: Vec<NamedAPIResource<Type>>,
    pub double_damage_to: Vec<NamedAPIResource<Type>>,
    pub no_damage_from: Vec<NamedAPIResource<Type>>,
    pub half_damage_from: Vec<NamedAPIResource<Type>>,
    pub double_damage_from: Vec<NamedAPIResource<Type>>,
}

#[derive(Serialize, Deserialize, Debug, Clone, PartialEq, Eq, Hash)]
#[cfg_attr(debug_assertions, serde(deny_unknown_fields))]
#[non_exhaustive]
pub struct PastTypeRelations {
    pub damage_relations: TypeRelations,
    pub generation: NamedAPIResource<Generation>,
}

set_endpoint!(Ability, NamedAPIResourceList, "ability");
set_endpoint!(Characteristic, APIResourceList, "characteristic");
set_endpoint!(EggGroup, NamedAPIResourceList, "egg-group");
set_endpoint!(Gender, NamedAPIResourceList, "gender");
set_endpoint!(GrowthRate, NamedAPIResourceList, "growth-rate");
set_endpoint!(Nature, NamedAPIResourceList, "nature");
set_endpoint!(PokeathlonStat, NamedAPIResourceList, "pokeathlon-stat");
set_endpoint!(Pokemon, NamedAPIResourceList, "pokemon");
set_endpoint!(PokemonColor, NamedAPIResourceList, "pokemon-color");
set_endpoint!(PokemonForm, NamedAPIResourceList, "pokemon-form");
set_endpoint!(PokemonHabitat, NamedAPIResourceList, "pokemon-habitat");
set_endpoint!(PokemonShape, NamedAPIResourceList, "pokemon-shape");
set_endpoint!(PokemonSpecies, NamedAPIResourceList, "pokemon-species");
set_endpoint!(Stat, NamedAPIResourceList, "stat");
set_endpoint!(Type, NamedAPIResourceList, "type");

impl_id!(Characteristic);
impl_id_and_named!(Ability);
impl_id_and_named!(EggGroup);
impl_id_and_named!(Gender);
impl_id_and_named!(GrowthRate);
impl_id_and_named!(Nature);
impl_id_and_named!(PokeathlonStat);
impl_id_and_named!(Pokemon);
impl_id_and_named!(PokemonColor);
impl_id_and_named!(PokemonForm);
impl_id_and_named!(PokemonHabitat);
impl_id_and_named!(PokemonShape);
impl_id_and_named!(PokemonSpecies);
impl_id_and_named!(Stat);
impl_id_and_named!(Type);
